from typing import Union
import os
import random
import string
import datetime
import shutil


class StagingArea:
    """
    A staging area where files can be stored temporarily before being
    consolidated and uploaded to a storage system.

    This class is a context manager, so it can be used in a `with` statement to
    ensure that the staging area is cleaned up when it is no longer needed.
    """
    def __init__(self, *, _directory: str) -> None:
        """
        Do not call this constructor directly. Instead, use the `create` method
        to create a new staging area.
        """
        self._directory = os.path.abspath(_directory)

    @staticmethod
    def create(*, base_dir: Union[str, None] = None, dir: Union[str, None] = None) -> 'StagingArea':
        """
        Create a new staging area. Provide either `base_dir` or `dir`, but not
        both.

        Parameters
        ----------
        base_dir : str or None
            If provided, the base directory where the staging area will be
            created. The staging directory will be a subdirectory of this
            directory.
        dir : str or None
            If provided, the exact directory where the staging area will be
            created. It is okay if this directory already exists.
        """
        if base_dir is not None and dir is not None:
            raise ValueError("Provide either base_dir or dir, but not both")
        if base_dir is not None:
            dir = os.path.join(base_dir, _create_random_id())
        if dir is None:
            raise ValueError("Provide either base_dir or dir")
        return StagingArea(_directory=dir)

    def cleanup(self) -> None:
        """
        Clean up the staging area, deleting all files in it. This method is
        called automatically when the staging area is used as a context manager
        in a `with` statement.
        """
        if os.path.exists(self._directory):
            shutil.rmtree(self._directory)

    def __enter__(self) -> 'StagingArea':
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.cleanup()

    @property
    def directory(self) -> str:
        """
        The directory where the files are stored.
        """
        return self._directory

    def store_file(self, relpath: str, value: bytes) -> str:
        """
        Store a file in the staging area.

        Parameters
        ----------
        relpath : str
            The relative path to the file, relative to the staging area root.
        value : bytes
            The contents of the file.
        """
        path = os.path.join(self._directory, relpath)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wb') as f:
            f.write(value)
        return path

    def get_full_path(self, relpath: str) -> str:
        """
        Get the full path to a file in the staging area.

        Parameters
        ----------
        relpath : str
            The relative path to the file, relative to the staging area root.
        """
        return os.path.join(self._directory, relpath)


def _create_random_id():
    # This is going to be a timestamp suitable for alphabetical chronological order plus a random string
    return f"{_timestamp_str()}-{_random_str(8)}"


def _timestamp_str():
    return datetime.datetime.now().strftime("%Y%m%d%H%M%S")


def _random_str(n):
    return ''.join(random.choices(string.ascii_lowercase + string.digits, k=n))



import h5py


class LindiH5pyReference(h5py.h5r.Reference):
    def __init__(self, obj: dict):
        self._obj = obj
        self._object_id = obj["object_id"]
        self._path = obj["path"]
        self._source = obj["source"]
        self._source_object_id = obj["source_object_id"]

    def __repr__(self):
        return f"LindiH5pyReference({self._object_id}, {self._path})"

    def __str__(self):
        return f"LindiH5pyReference({self._object_id}, {self._path})"



from typing import Union
import time
import os
import requests
from .additional_url_resolvers import get_additional_url_resolvers
from ..LocalCache.LocalCache import LocalCache

default_min_chunk_size = 128 * 1024  # This is different from Remfile - this is an important decision because it determines the chunk size in the LocalCache
default_max_cache_size = 1024 * 1024 * 1024
default_chunk_increment_factor = 1.7
default_max_chunk_size = 100 * 1024 * 1024


class LindiRemfile:
    def __init__(
        self,
        url: str,
        *,
        verbose: bool = False,
        local_cache: Union[LocalCache, None],
        _min_chunk_size: int = default_min_chunk_size,  # recommended not to change because it will make the accumulated cache in LocalCache useless
        _max_cache_size: int = default_max_cache_size,
        _chunk_increment_factor: float = default_chunk_increment_factor,
        _max_chunk_size: int = default_max_cache_size,
        _impose_request_failures_for_testing: bool = False,
    ):
        """Create a file-like object for reading a remote file. Optimized for reading hdf5 files. The arguments starting with an underscore are for testing and debugging purposes - they may experience breaking changes in the future.

        Args:
            url (str): The url of the remote file, or an object with a .get_url() method. The latter is useful if the url is a presigned AWS URL that expires after a certain amount of time.
            verbose (bool, optional): Whether to print info for debugging. Defaults to False.
            local_cache (LocalCache, optional): The local cache for storing the chunks of the file. Defaults to None.
            _min_chunk_size (int, optional): The minimum chunk size. When reading, the chunks will be loaded in multiples of this size.
            _max_cache_size (int, optional): The maximum number of bytes to keep in the cache.
            _chunk_increment_factor (int, optional): The factor by which to increase the number of chunks to load when the system detects that the chunks are being loaded in order.
            _max_chunk_size (int, optional): The maximum chunk size. When reading, the chunks will be loaded in multiples of the minimum chunk size up to this size.
            _impose_request_failures_for_testing (bool, optional): Whether to impose request failures for testing purposes. Defaults to False.

        Difference between this and regular RemFile?
            Uses Lindi's LocalCache instead of Remfile's DiskCache
            Requires that url is a string (does not accept object with .get_url() function)
            Does not support using multiple threads
            Does not use memory cache if LocalCache is specified
            Handles DANDI authentication
            Handles additional_url_resolver

        A note:
            In the context of LINDI, this LindiRemfile is going to be used for loading
            metadata of the hdf5 file. The large chunks are going to be loaded using
            a different (Zarr) mechanism. That's one reason for the above differences.
        """
        if not isinstance(url, str):
            # Do this check, because regular Remfile allows for objects
            raise Exception('Only string urls are supported for LindiRemfile')
        self._url = url
        self._verbose = verbose
        self._local_cache = local_cache
        self._min_chunk_size = _min_chunk_size
        self._max_chunks_in_cache = int(_max_cache_size / _min_chunk_size)
        self._chunk_increment_factor = _chunk_increment_factor
        self._max_chunk_size = _max_chunk_size
        self._impose_request_failures_for_testing = _impose_request_failures_for_testing
        self._memory_chunks = {}
        self._memory_chunk_indices: list[
            int
        ] = (
            []
        )  # list of chunk indices in order of loading for purposes of cleaning up the cache
        self._position = 0
        self._smart_loader_last_chunk_index_accessed = -99
        self._smart_loader_chunk_sequence_length = 1

        # use aborted GET request rather than HEAD request to get the length
        # this is needed for presigned AWS URLs because HEAD requests are not supported
        response = requests.get(_resolve_url(self._url), stream=True)
        if response.status_code == 200:
            self.length = int(response.headers["Content-Length"])
        else:
            raise Exception(
                f"Error getting file length: {response.status_code} {response.reason}"
            )
        # Close the connection without reading the content to avoid downloading the whole file
        response.close()

        # response = requests.head(_get_url_str(self._url))
        # self.length = int(response.headers['Content-Length'])
        self.session = requests.Session()

    def read(self, size=None):
        """Read bytes from the file.

        Args:
            size (_type_): The number of bytes to read.

        Raises:
            Exception: If the size argument is not provided.

        Returns:
            bytes: The bytes read.
        """
        if size is None:
            raise Exception(
                "The size argument must be provided in remfile"
            )  # pragma: no cover

        chunk_start_index = self._position // self._min_chunk_size
        chunk_end_index = (self._position + size - 1) // self._min_chunk_size
        loaded_chunks = {}
        for chunk_index in range(chunk_start_index, chunk_end_index + 1):
            loaded_chunks[chunk_index] = self._load_chunk(chunk_index)
        if chunk_end_index == chunk_start_index:
            chunk = loaded_chunks[chunk_start_index]
            chunk_offset = self._position % self._min_chunk_size
            chunk_length = size
            self._position += size
            return chunk[chunk_offset: chunk_offset + chunk_length]
        else:
            pieces_to_concat = []
            for chunk_index in range(chunk_start_index, chunk_end_index + 1):
                chunk = loaded_chunks[chunk_index]
                if chunk_index == chunk_start_index:
                    chunk_offset = self._position % self._min_chunk_size
                    chunk_length = self._min_chunk_size - chunk_offset
                elif chunk_index == chunk_end_index:
                    chunk_offset = 0
                    chunk_length = size - sum([len(p) for p in pieces_to_concat])
                else:
                    chunk_offset = 0
                    chunk_length = self._min_chunk_size
                pieces_to_concat.append(
                    chunk[chunk_offset: chunk_offset + chunk_length]
                )
        ret = b"".join(pieces_to_concat)
        self._position += size

        # clean up the cache
        if len(self._memory_chunk_indices) > self._max_chunks_in_cache:
            if self._verbose:
                print("Cleaning up cache")
            for chunk_index in self._memory_chunk_indices[
                : int(self._max_chunks_in_cache * 0.5)
            ]:
                if chunk_index in self._memory_chunks:
                    del self._memory_chunks[chunk_index]
                else:
                    # it is possible that the chunk was already deleted (repeated chunk index in the list)
                    pass
            self._memory_chunk_indices = self._memory_chunk_indices[
                int(self._max_chunks_in_cache * 0.5):
            ]

        return ret

    def _load_chunk(self, chunk_index: int) -> bytes:
        """Load a chunk of the file.

        Args:
            chunk_index (int): The index of the chunk to load.
        """
        if chunk_index in self._memory_chunks:
            self._smart_loader_last_chunk_index_accessed = chunk_index
            return self._memory_chunks[chunk_index]

        if self._local_cache is not None:
            cached_value = self._local_cache.get_remote_chunk(
                url=self._url,
                offset=chunk_index * self._min_chunk_size,
                size=min(self._min_chunk_size, self.length - chunk_index * self._min_chunk_size),
            )
            if cached_value is not None:
                if self._local_cache is None:
                    self._memory_chunks[chunk_index] = cached_value
                    self._memory_chunk_indices.append(chunk_index)
                self._smart_loader_last_chunk_index_accessed = chunk_index
                return cached_value

        if chunk_index == self._smart_loader_last_chunk_index_accessed + 1:
            # round up to the chunk sequence length times 1.7
            self._smart_loader_chunk_sequence_length = round(
                self._smart_loader_chunk_sequence_length * 1.7 + 0.5
            )
            if (
                self._smart_loader_chunk_sequence_length > self._max_chunk_size / self._min_chunk_size
            ):
                self._smart_loader_chunk_sequence_length = int(
                    self._max_chunk_size / self._min_chunk_size
                )
            # make sure the chunk sequence length is valid
            for j in range(1, self._smart_loader_chunk_sequence_length):
                if chunk_index + j in self._memory_chunks:
                    # already loaded this chunk
                    self._smart_loader_chunk_sequence_length = j
                    break
        else:
            self._smart_loader_chunk_sequence_length = round(
                self._smart_loader_chunk_sequence_length / 1.7 + 0.5
            )
        data_start = chunk_index * self._min_chunk_size
        data_end = (
            data_start + self._min_chunk_size * self._smart_loader_chunk_sequence_length - 1
        )
        if self._verbose:
            print(
                f"Loading {self._smart_loader_chunk_sequence_length} chunks starting at {chunk_index} ({(data_end - data_start + 1)/1e6} million bytes)"
            )
        if data_end >= self.length:
            data_end = self.length - 1
        x = _get_bytes(
            self.session,
            _resolve_url(self._url),
            data_start,
            data_end,
            verbose=self._verbose,
            _impose_request_failures_for_testing=self._impose_request_failures_for_testing,
        )
        if not x:
            raise Exception(f'Error loading chunk {chunk_index} from {self._url}')
        if self._smart_loader_chunk_sequence_length == 1:
            if self._local_cache is None:
                self._memory_chunks[chunk_index] = x
            if self._local_cache is not None:
                size = min(self._min_chunk_size, self.length - chunk_index * self._min_chunk_size)
                self._local_cache.put_remote_chunk(
                    url=self._url,
                    offset=chunk_index * self._min_chunk_size,
                    size=size,
                    data=x
                )
            self._memory_chunk_indices.append(chunk_index)
        else:
            for i in range(self._smart_loader_chunk_sequence_length):
                if i * self._min_chunk_size >= len(x):
                    break
                if self._local_cache is None:
                    self._memory_chunks[chunk_index + i] = x[
                        i * self._min_chunk_size: (i + 1) * self._min_chunk_size
                    ]
                    self._memory_chunk_indices.append(chunk_index + i)
                if self._local_cache is not None:
                    size = min(self._min_chunk_size, self.length - (chunk_index + i) * self._min_chunk_size)
                    data = x[i * self._min_chunk_size: (i + 1) * self._min_chunk_size]
                    if len(data) != size:
                        raise ValueError(f'Unexpected: len(data) != size: {len(data)} != {size}')
                    self._local_cache.put_remote_chunk(
                        url=self._url,
                        offset=(chunk_index + i) * self._min_chunk_size,
                        size=size,
                        data=data
                    )
        self._smart_loader_last_chunk_index_accessed = (
            chunk_index + self._smart_loader_chunk_sequence_length - 1
        )
        return x[: self._min_chunk_size]

    def seek(self, offset: int, whence: int = 0):
        """Seek to a position in the file.

        Args:
            offset (int): The offset to seek to.
            whence (int, optional): The code for the reference point for the offset. Defaults to 0.

        Raises:
            ValueError: If the whence argument is not 0, 1, or 2.
        """
        if whence == 0:
            self._position = offset
        elif whence == 1:
            self._position += offset  # pragma: no cover
        elif whence == 2:
            self._position = self.length + offset
        else:
            raise ValueError(
                "Invalid argument: 'whence' must be 0, 1, or 2."
            )  # pragma: no cover

    def tell(self):
        return self._position

    def close(self):
        pass


_num_request_retries = 8


def _get_bytes(
    session: requests.Session,
    url: str,
    start_byte: int,
    end_byte: int,
    *,
    verbose=False,
    _impose_request_failures_for_testing=False,
):
    """Get bytes from a remote file.

    Args:
        url (str): The url of the remote file.
        start_byte (int): The first byte to get.
        end_byte (int): The last byte to get.
        verbose (bool, optional): Whether to print info for debugging. Defaults to False.

    Returns:
        _type_: _description_
    """
    # Function to be used in threads for fetching the byte ranges
    def fetch_bytes(range_start: int, range_end: int, num_retries: int, verbose: bool):
        """Fetch a range of bytes from a remote file using the range header

        Args:
            range_start (int): The first byte to get.
            range_end (int): The last byte to get.
            num_retries (int): The number of retries.

        Returns:
            bytes: The bytes fetched.
        """
        for try_num in range(num_retries + 1):
            try:
                actual_url = url
                if _impose_request_failures_for_testing:
                    if try_num == 0:
                        actual_url = "_error_" + url
                range_header = f"bytes={range_start}-{range_end}"
                # response = requests.get(actual_url, headers={'Range': range_header})
                # use session to avoid creating a new connection each time
                response = session.get(actual_url, headers={"Range": range_header})
                return response.content
            except Exception as e:
                if try_num == num_retries:
                    raise e  # pragma: no cover
                else:
                    delay = 0.1 * 2**try_num
                    if verbose:
                        print(f"Retrying after exception: {e}")
                        print(f"Waiting {delay} seconds")
                    time.sleep(delay)

    return fetch_bytes(start_byte, end_byte, _num_request_retries, verbose)


_global_resolved_urls = {}  # url -> {timestamp, url}


def _is_dandi_url(url: str):
    if url.startswith('https://api.dandiarchive.org/api/'):
        return True
    if url.startswith('https://api-staging.dandiarchive.org/'):
        return True
    return False


def _resolve_dandi_url(url: str):
    resolve_with_dandi_api_key = None
    if url.startswith('https://api.dandiarchive.org/api/'):
        dandi_api_key = os.environ.get('DANDI_API_KEY', None)
        if dandi_api_key is not None:
            resolve_with_dandi_api_key = dandi_api_key
    elif url.startswith('https://api-staging.dandiarchive.org/'):
        dandi_api_key = os.environ.get('DANDI_STAGING_API_KEY', None)
        if dandi_api_key is not None:
            resolve_with_dandi_api_key = dandi_api_key
    headers = {}
    if resolve_with_dandi_api_key is not None:
        headers['Authorization'] = f'token {resolve_with_dandi_api_key}'
    # do it synchronously here
    resp = requests.head(url, allow_redirects=True, headers=headers)
    return str(resp.url)


def _resolve_url(url: str):
    for aur in get_additional_url_resolvers():
        url = aur(url)
    if url in _global_resolved_urls:
        elapsed = time.time() - _global_resolved_urls[url]["timestamp"]
        if elapsed < 60 * 10:
            return _global_resolved_urls[url]["url"]
    if _is_dandi_url(url):
        resolved_url = _resolve_dandi_url(url)
    else:
        resolved_url = url
    _global_resolved_urls[url] = {"timestamp": time.time(), "url": resolved_url}
    return resolved_url



from typing import Union
from dataclasses import dataclass


@dataclass(frozen=True)
class LindiH5ZarrStoreOpts:
    """
    Options for the LindiH5ZarrStore class.

    Attributes:
        num_dataset_chunks_threshold (Union[int, None]): For each dataset in the
        HDF5 file, if the number of chunks is greater than this threshold, then
        the dataset will be represented as an external array link. If None, then
        no datasets will be represented as external array links (equivalent to a
        threshold of 0). Default is 1000.

        contiguous_dataset_max_chunk_size (Union[int, None]): For large
        contiguous arrays in the hdf5 file that are not chunked, this option
        specifies the maximum size in bytes of the zarr chunks that will be
        created. If None, then the entire array will be represented as a single
        chunk. Default is 1000 * 1000 * 20
    """
    num_dataset_chunks_threshold: Union[int, None] = 1000
    contiguous_dataset_max_chunk_size: Union[int, None] = 1000 * 1000 * 20



from typing import Any, Union
import numpy as np
import h5py
from .nan_inf_ninf import encode_nan_inf_ninf
from .h5_ref_to_zarr_attr import h5_ref_to_zarr_attr


def h5_to_zarr_attr(attr: Any, *, label: str = '', h5f: Union[h5py.File, None]):
    """Convert an attribute from h5py to a format that zarr can accept."""

    from ..LindiH5pyFile.LindiH5pyReference import LindiH5pyReference  # Avoid circular import

    # Do not allow these special strings in attributes
    special_strings = ['NaN', 'Infinity', '-Infinity']

    if isinstance(attr, list):
        list_dtype = _determine_list_dtype(attr)
        attr = np.array(attr, dtype=list_dtype)

    if isinstance(attr, str) and attr in special_strings:
        raise ValueError(f"Special string {attr} not allowed in attribute value at {label}")
    if attr is None:
        raise Exception(f"Unexpected h5 attribute: None at {label}")
    elif type(attr) in [int, np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64]:
        return int(attr)
    elif isinstance(attr, (float, np.floating)):
        return encode_nan_inf_ninf(float(attr))
    elif isinstance(attr, (complex, np.complexfloating)):
        raise Exception(f"Complex number is not supported at {label}")
    elif type(attr) in [bool, np.bool_]:
        return bool(attr)
    elif isinstance(attr, (bool, list, tuple, dict, set)):
        raise Exception(f"Unexpected type for h5 attribute: {type(attr)} at {label}")
    elif isinstance(attr, str):
        return attr
    elif isinstance(attr, bytes):
        return attr.decode('utf-8')
    elif isinstance(attr, np.ndarray):
        if attr.dtype.kind in ['i', 'u']:
            return attr.tolist()  # this will be a nested list of type int
        elif attr.dtype.kind in ['f']:
            return encode_nan_inf_ninf(attr.tolist())  # this will be a nested list of type float
        elif attr.dtype.kind in ['c']:
            raise Exception(f"Arrays of complex numbers are not supported at {label}")
        elif attr.dtype.kind == 'b':
            return attr.tolist()  # this will be a nested list of type bool
        elif attr.dtype.kind == 'O':
            x = attr.tolist()
            if not _nested_list_has_all_strings(x):
                raise Exception(f"Not allowed for attribute: numpy array with dtype=object that contains non-string elements at {label}")
            return x
        elif attr.dtype.kind == 'U':
            return _decode_bytes_to_str_in_nested_list(attr.tolist())
        elif attr.dtype.kind == 'S':
            return _decode_bytes_to_str_in_nested_list(attr.tolist())
        else:
            raise Exception(f"Unexpected dtype for attribute numpy array: {attr.dtype} at {label}")
    elif isinstance(attr, LindiH5pyReference):
        return {
            '_REFERENCE': attr._obj
        }
    elif isinstance(attr, h5py.Reference):
        if h5f is None:
            raise Exception(f"h5f cannot be None when converting h5py.Reference to zarr attribute at {label}")
        return h5_ref_to_zarr_attr(attr, h5f=h5f)
    else:
        raise Exception(f"Unexpected type for h5 attribute: {type(attr)} at {label}")


def _decode_bytes_to_str_in_nested_list(x):
    if isinstance(x, bytes):
        return x.decode('utf-8')
    elif isinstance(x, str):
        return x
    elif isinstance(x, list):
        return [_decode_bytes_to_str_in_nested_list(y) for y in x]
    else:
        raise Exception("Unexpected type in _decode_bytes_to_str")


def zarr_to_h5_attr(attr: Any):
    """Convert an attribute from zarr to a format that h5py expects."""
    if isinstance(attr, str):
        return attr
    elif isinstance(attr, int):
        return attr
    elif isinstance(attr, float):
        return attr
    elif isinstance(attr, bool):
        return np.bool_(attr)
    elif isinstance(attr, list):
        if _nested_list_has_all_strings(attr):
            return np.array(attr, dtype='O')
        elif _nested_list_has_all_ints(attr):
            return np.array(attr, dtype='int64')
        elif _nested_list_has_all_floats_or_ints(attr):
            return np.array(attr, dtype='float64')
        elif _nested_list_has_all_bools(attr):
            return np.array(attr, dtype='bool')
        else:
            raise Exception("Nested list contains mixed types")
    else:
        raise Exception(f"Unexpected type in zarr attribute: {type(attr)}")


def _nested_list_has_all_strings(x):
    if isinstance(x, str):
        return True
    elif isinstance(x, list):
        return all(_nested_list_has_all_strings(y) for y in x)
    else:
        return False


def _nested_list_has_all_ints(x):
    if isinstance(x, int):
        return True
    elif isinstance(x, list):
        return all(_nested_list_has_all_ints(y) for y in x)
    else:
        return False


def _nested_list_has_all_floats_or_ints(x):
    if isinstance(x, (int, float)):
        return True
    elif isinstance(x, list):
        return all(_nested_list_has_all_floats_or_ints(y) for y in x)
    else:
        return False


def _nested_list_has_all_bools(x):
    if isinstance(x, bool):
        return True
    elif isinstance(x, list):
        return all(_nested_list_has_all_bools(y) for y in x)
    else:
        return False


def _determine_list_dtype(x):
    x_flattened = _flatten_list(x)
    if len(x_flattened) == 0:
        return np.dtype(np.int64)
    if all(isinstance(i, int) for i in x_flattened):
        return np.int64
    elif all(isinstance(i, float) for i in x_flattened):
        return np.float64
    elif all(isinstance(i, bool) for i in x_flattened):
        return np.bool_
    elif all(isinstance(i, str) for i in x_flattened):
        return np.dtype('O')
    else:
        raise Exception("Mixed types in list")


def _flatten_list(x):
    if isinstance(x, list):
        return [a for i in x for a in _flatten_list(i)]
    else:
        return [x]



from typing import Callable, List


_global = {
    'additional_url_resolvers': []
}


def get_additional_url_resolvers() -> List[Callable[[str], str]]:
    return _global['additional_url_resolvers']


def add_additional_url_resolver(resolver: Callable[[str], str]):
    _global['additional_url_resolvers'].append(resolver)



from typing import Literal, Dict, Union
import os
import json
import time
import base64
import numpy as np
import requests
from zarr.storage import Store as ZarrStore

from ..LocalCache.LocalCache import ChunkTooLargeError, LocalCache
from ..tar.lindi_tar import LindiTarFile


class LindiReferenceFileSystemStore(ZarrStore):
    """
    A Zarr store that reads data from a reference file system.

    The reference file system is based on the ReferenceFileSystem of fsspec, but
    this is a custom implementation that serves some LINDI-specific needs. In
    particular, it handles reading data from DANDI URLs, even when the file is
    part of an embargoed dataset. This requires some special handling as the
    DANDI API URL must be exchanged for a pre-signed S3 bucket URL by
    authenticating with a DANDI API token. This presigned URL expires after a
    period of time, so this Zarr store handles the renewal of the presigned URL.
    It also does the exchange once the first time and caches the redirected URL
    for a period so that the redirect doesn't need to be done every time a
    segment of a file is read.

    To read from a file in an embargoed DANDI dataset, you will need to set the
    DANDI_API_KEY environment variable to your DANDI API token. Or, if this is a
    Dandiset on the staging server, you will need to set the
    DANDI_STAGING_API_KEY.

    Following the fsspec convention
    (https://fsspec.github.io/kerchunk/spec.html), the reference file system is
    specified as a dictionary with a "refs" key. The value of "refs" is a
    dictionary where the keys are the names of the files and each value is
    either a string, a list, or a dict. If the value is a string, it is assumed
    to be the data of the file, which may be base64 encoded (see below). If the
    value is a list, it is assumed to have three elements: the URL of the file
    (or path of a local file), the byte offset of the data within the file, and
    the byte length of the data. Note that we do not permit the case of a list
    of a single (url) element supported by fsspec, because it is good to be able
    to know the size of the chunks without making a request to the file. If the
    value is a dict, it represents a json file, and the content of the file is
    the json representation of the dict.

    If the value for a file is a string, it may be prefixed with "base64:". If
    it is, the string is assumed to be base64 encoded and is decoded before
    being returned. Otherwise, the string is utf-8 encoded and returned as is.
    Note that a file that actually begins with "base64:" should be represented
    by a base64 encoded string, to avoid ambiguity.

    We also support the use of templates as in fsspec, but do not support the
    full jinja2 templating. There may be an optional "templates" key in the
    dictionary, which is a dictionary of template strings. For example, {
        "templates": {"u1": "https://some/url", "u2": "https://some/other/url"},
        "refs": {
            ... "/some/key/0": [
                "{{u1}}" 0, 100
            ],
            ...
        }
    } In this case, the "{{u1}}" will be replaced with the value of the "u1"
    template string.

    Optionally, the reference file system may contain a "generationMetadata"
    key, which is a dictionary of metadata about the generation of the reference
    file system. This metadata is not used by this class, but could be by other
    software. See LindiH5pyFile.write_lindi_file(...)

    It is okay for rfs to be modified outside of this class, and the changes
    will be reflected immediately in the store.
    """
    def __init__(
            self,
            rfs: dict,
            *,
            mode: Literal["r", "r+"] = "r+",
            local_cache: Union[LocalCache, None] = None,
            _source_url_or_path: Union[str, None] = None,
            _source_tar_file: Union[LindiTarFile, None] = None
    ):
        """
        Create a LindiReferenceFileSystemStore.

        Parameters
        ----------
        rfs : dict
            The reference file system (see class docstring for details).
        mode : str
            The mode to open the store in. Only "r" is supported at this time.
        local_cache : LocalCache, optional
            The local cache to use for caching data chunks read from the
            remote URLs. If None, no caching is done.
        """
        if "refs" not in rfs:
            raise Exception("rfs must contain a 'refs' key")

        # validate rfs['refs']
        for k, v in rfs["refs"].items():
            if isinstance(v, str):
                pass
            elif isinstance(v, dict):
                # the content of the file is the json representation of the dict
                pass
            elif isinstance(v, list):
                if len(v) != 3:
                    raise Exception(f"Problem with {k}: list must have 3 elements")
                if not isinstance(v[0], str):
                    raise Exception(f"Problem with {k}: first element must be a string")
                if not isinstance(v[1], int):
                    raise Exception(f"Problem with {k}: second element must be an int")
                if not isinstance(v[2], int):
                    raise Exception(f"Problem with {k}: third element must be an int")
            else:
                raise Exception(f"Problem with {k}: value must be a string or a list")

        # validate templates
        if "templates" in rfs:
            for k, v in rfs["templates"].items():
                if not isinstance(v, str):
                    raise Exception(f"Problem with templates: value for {k} must be a string")

        self.rfs = rfs
        self.mode = mode
        self.local_cache = local_cache
        self._source_url_or_path = _source_url_or_path
        self._source_tar_file = _source_tar_file

    # These methods are overridden from MutableMapping
    def __contains__(self, key: object):
        if not isinstance(key, str):
            return False
        return key in self.rfs["refs"]

    def __getitem__(self, key: str):
        val = self._get_helper(key)

        if val is not None:
            padded_size = _get_padded_size(self, key, val)
            if padded_size is not None:
                val = _pad_chunk(val, padded_size)

        return val

    def _get_helper(self, key: str):
        if key not in self.rfs["refs"]:
            raise KeyError(key)
        x = self.rfs["refs"][key]
        if isinstance(x, str):
            if x.startswith("base64:"):
                return base64.b64decode(x[len("base64:"):])
            else:
                return x.encode("utf-8")
        elif isinstance(x, dict):
            return json.dumps(x).encode("utf-8")
        elif isinstance(x, list):
            if len(x) != 3:
                raise Exception("list must have 3 elements")  # pragma: no cover
            url_or_path = x[0]
            offset = x[1]
            length = x[2]
            if '{{' in url_or_path and '}}' in url_or_path and 'templates' in self.rfs:
                for k, v in self.rfs["templates"].items():
                    url_or_path = url_or_path.replace("{{" + k + "}}", v)
            is_url = url_or_path.startswith('http://') or url_or_path.startswith('https://')
            if url_or_path.startswith('./'):
                if self._source_url_or_path is None:
                    raise Exception(f"Cannot resolve relative path {url_or_path} without source file path")
                if self._source_tar_file is None:
                    raise Exception(f"Cannot resolve relative path {url_or_path} without source file type")
                if self._source_tar_file and (not self._source_tar_file._dir_representation):
                    start_byte, end_byte = self._source_tar_file.get_file_byte_range(file_name=url_or_path[2:])
                    if start_byte + offset + length > end_byte:
                        raise Exception(f"Chunk {key} is out of bounds in tar file {url_or_path}")
                    url_or_path = self._source_url_or_path
                    offset = offset + start_byte
                elif self._source_tar_file and self._source_tar_file._dir_representation:
                    fname = self._source_tar_file._tar_path_or_url + '/' + url_or_path[2:]
                    if not os.path.exists(fname):
                        raise Exception(f"File does not exist: {fname}")
                    file_size = os.path.getsize(fname)
                    if offset + length > file_size:
                        raise Exception(f"Chunk {key} is out of bounds in tar file {url_or_path}: {fname}")
                    url_or_path = fname
                else:
                    if is_url:
                        raise Exception(f"Cannot resolve relative path {url_or_path} for URL that is not a tar")
                    else:
                        source_file_parent_dir = '/'.join(self._source_url_or_path.split('/')[:-1])
                        abs_path = source_file_parent_dir + '/' + url_or_path[2:]
                        url_or_path = abs_path
            if is_url:
                if self.local_cache is not None:
                    x = self.local_cache.get_remote_chunk(url=url_or_path, offset=offset, size=length)
                    if x is not None:
                        return x
                val = _read_bytes_from_url_or_path(url_or_path, offset, length)
                if self.local_cache is not None:
                    try:
                        self.local_cache.put_remote_chunk(url=url_or_path, offset=offset, size=length, data=val)
                    except ChunkTooLargeError:
                        print(f'Warning: unable to cache chunk of size {length} on LocalCache (key: {key})')
            else:
                val = _read_bytes_from_url_or_path(url_or_path, offset, length)
            return val
        else:
            # should not happen given checks in __init__, but self.rfs is mutable
            # and contains mutable lists
            raise Exception(f"Problem with {key}: value {x} must be a string or a list")

    def __setitem__(self, key: str, value: bytes):
        # We intentionally do not allow value to be a dict here! When the rfs is
        # written to a .json file elsewhere in the codebase of lindi, the value
        # will automatically be converted to a json object if it is json
        # serializable.
        if not isinstance(value, bytes):
            raise ValueError("value must be bytes")
        try:
            # try to ascii encode the value
            value2 = value.decode("ascii")
        except UnicodeDecodeError:
            # if that fails, base64 encode it
            value2 = "base64:" + base64.b64encode(value).decode("ascii")
        self.rfs["refs"][key] = value2

    def __delitem__(self, key: str):
        del self.rfs["refs"][key]

    def __iter__(self):
        return iter(self.rfs["refs"])

    def __len__(self):
        return len(self.rfs["refs"])

    # These methods are overridden from BaseStore
    def is_readable(self):
        return self.mode in ["r", "r+"]

    def is_writeable(self):
        return self.mode in ["r+"]

    def is_listable(self):
        return True

    def is_erasable(self):
        return False

    @staticmethod
    def replace_meta_file_contents_with_dicts_in_rfs(rfs: dict) -> None:
        """
        Utility function for replacing the contents of the .zattrs, .zgroup, and
        .zarray files in an rfs with the json representation of the contents.
        """
        # important to use the LindiReferenceFileSystemStore here because then we
        # can resolve any base64 encoded values, etc when converting them to dicts
        store = LindiReferenceFileSystemStore(rfs)
        for k, v in rfs['refs'].items():
            if k.endswith('.zattrs') or k.endswith('.zgroup') or k.endswith('.zarray') or k.endswith('zarr.json'):  # note: zarr.json is for zarr v3
                rfs['refs'][k] = json.loads(store[k].decode('utf-8'))

    @staticmethod
    def use_templates_in_rfs(rfs: dict) -> None:
        """
        Utility for replacing URLs in an rfs with template strings. Only URLs
        that occur 5 or more times are replaced with template strings. The
        templates are added to the "templates" key of the rfs. The template
        strings are of the form "{{u1}}", "{{u2}}", etc.
        """
        url_counts: Dict[str, int] = {}
        for k, v in rfs['refs'].items():
            if isinstance(v, list):
                url = v[0]
                if '{{' not in url:
                    url_counts[url] = url_counts.get(url, 0) + 1
        urls_with_many_occurrences = sorted([url for url, count in url_counts.items() if count >= 5])
        new_templates = rfs.get('templates', {})
        template_names_for_urls: Dict[str, str] = {}
        for template_name, url in new_templates.items():
            template_names_for_urls[url] = template_name
        for url in urls_with_many_occurrences:
            if url in template_names_for_urls:
                continue
            i = 1
            while f'u{i}' in new_templates:
                i += 1
            new_templates[f'u{i}'] = url
            template_names_for_urls[url] = f'u{i}'
        if new_templates:
            rfs['templates'] = new_templates
        for k, v in rfs['refs'].items():
            if isinstance(v, list):
                url = v[0]
                if url in template_names_for_urls:
                    v[0] = '{{' + template_names_for_urls[url] + '}}'

    @staticmethod
    def remove_templates_in_rfs(rfs: dict) -> None:
        """
        Utility for removing templates from an rfs. This is the opposite of
        use_templates_in_rfs.
        """
        templates0 = rfs.get('templates', {})
        for k, v in rfs['refs'].items():
            if isinstance(v, list):
                url = v[0]
                if '{{' in url and '}}' in url:
                    template_name = url[2:-2].strip()
                    if template_name in templates0:
                        v[0] = templates0[template_name]
        rfs['templates'] = {}


def _read_bytes_from_url_or_path(url_or_path: str, offset: int, length: int):
    """
    Read a range of bytes from a URL.
    """
    from ..LindiRemfile.LindiRemfile import _resolve_url
    if url_or_path.startswith('http://') or url_or_path.startswith('https://'):
        num_retries = 8
        for try_num in range(num_retries):
            try:
                url_resolved = _resolve_url(url_or_path)  # handle DANDI auth
                range_start = offset
                range_end = offset + length - 1
                range_header = f"bytes={range_start}-{range_end}"
                headers = {
                    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3",
                    "Range": range_header
                }
                response = requests.get(url_resolved, headers=headers)
                response.raise_for_status()
                return response.content
            except Exception as e:
                if try_num == num_retries - 1:
                    raise e
                else:
                    delay = 0.1 * 2 ** try_num
                    print(f'Retry load data from {url_or_path} in {delay} seconds')
                    time.sleep(delay)
        raise Exception(f"Failed to load data from {url_or_path}")
    else:
        with open(url_or_path, 'rb') as f:
            f.seek(offset)
            return f.read(length)


def _is_chunk_base_key(base_key: str) -> bool:
    a = base_key.split('.')
    if len(a) == 0:
        return False
    for x in a:
        # check if integer
        try:
            int(x)
        except ValueError:
            return False
    return True


def _get_itemsize(dtype: str) -> int:
    d = np.dtype(dtype)
    return d.itemsize


def _pad_chunk(data: bytes, expected_chunk_size: int) -> bytes:
    return data + b'\0' * (expected_chunk_size - len(data))


def _get_padded_size(store, key: str, val: bytes):
    # If the key is a chunk and it's smaller than the expected size, then we
    # need to pad it with zeros. This can happen if this is the final chunk
    # in a contiguous hdf5 dataset. See
    # https://github.com/NeurodataWithoutBorders/lindi/pull/84
    base_key = key.split('/')[-1]
    if val and _is_chunk_base_key(base_key):
        parent_key = key.split('/')[:-1]
        zarray_key = '/'.join(parent_key) + '/.zarray'
        if zarray_key in store:
            zarray_json = store.__getitem__(zarray_key)
            assert isinstance(zarray_json, bytes)
            zarray = json.loads(zarray_json)
            chunk_shape = zarray['chunks']
            dtype = zarray['dtype']
            if np.dtype(dtype).kind in ['i', 'u', 'f']:
                expected_chunk_size = int(np.prod(chunk_shape)) * _get_itemsize(dtype)
                if len(val) < expected_chunk_size:
                    return expected_chunk_size

    return None



def create_tar_header(file_name: str, file_size: int) -> bytes:
    # We use USTAR format only
    h = b''

    # file name
    a = file_name.encode() + b"\x00" * (100 - len(file_name))
    h += a

    # file mode
    a = b"0000644\x00"  # 644 is the default permission - you can read and write, but others can only read
    h += a

    # uid
    a = b"0000000\x00"  # 0 is the default user id
    h += a

    # gid
    a = b"0000000\x00"  # 0 is the default group id
    h += a

    # size
    # we need an octal representation of the size
    a = f"{file_size:011o}".encode() + b"\x00"  # 11 octal digits
    h += a

    # mtime
    a = b"00000000000\x00"  # 0 is the default modification time
    h += a

    # chksum
    # We'll determine the checksum after creating the full header
    a = b" " * 8  # 8 spaces for now
    h += a

    # typeflag
    a = b"0"  # default typeflag is 0 representing a regular file
    h += a

    # linkname
    a = b"\x00" * 100  # no link name
    h += a

    # magic
    a = b"ustar\x00"  # specifies the ustar format
    h += a

    # version
    a = b"00"  # ustar version
    h += a

    # uname
    a = b"\x00" * 32  # no user name
    h += a

    # gname
    a = b"\x00" * 32  # no group name
    h += a

    # devmajor
    a = b"\x00" * 8  # no device major number
    h += a

    # devminor
    a = b"\x00" * 8  # no device minor number
    h += a

    # prefix
    a = b"\x00" * 155  # no prefix
    h += a

    # padding
    a = b"\x00" * 12  # padding
    h += a

    # Now we calculate the checksum
    chksum = _compute_checksum_for_header(h)
    h = h[:148] + chksum + h[156:]

    assert len(h) == 512

    return h


def _compute_checksum_for_header(header: bytes) -> bytes:
    # From https://en.wikipedia.org/wiki/Tar_(computing)
    # The checksum is calculated by taking the sum of the unsigned byte values
    # of the header record with the eight checksum bytes taken to be ASCII
    # spaces (decimal value 32). It is stored as a six digit octal number with
    # leading zeroes followed by a NUL and then a space. Various implementations
    # do not adhere to this format. In addition, some historic tar
    # implementations treated bytes as signed. Implementations typically
    # calculate the checksum both ways, and treat it as good if either the
    # signed or unsigned sum matches the included checksum.

    header_byte_list = []
    for byte in header:
        header_byte_list.append(byte)
    for i in range(148, 156):
        header_byte_list[i] = 32
    sum = 0
    for byte in header_byte_list:
        sum += byte
    checksum = oct(sum).encode()[2:]
    while len(checksum) < 6:
        checksum = b"0" + checksum
    checksum += b"\0 "
    return checksum



from .LindiH5ZarrStore import LindiH5ZarrStore, LindiH5ZarrStoreOpts  # noqa: F401
from .LindiH5pyFile import LindiH5pyFile, LindiH5pyGroup, LindiH5pyDataset, LindiH5pyHardLink, LindiH5pySoftLink  # noqa: F401
from .LindiStagingStore import LindiStagingStore, StagingArea  # noqa: F401
from .LocalCache.LocalCache import LocalCache, ChunkTooLargeError  # noqa: F401
from .File.File import File  # noqa: F401
from .LindiRemfile.additional_url_resolvers import add_additional_url_resolver  # noqa: F401



from typing import Union, List, Any, Tuple, Literal
from dataclasses import dataclass
import numpy as np
import numcodecs
from numcodecs.abc import Codec
import h5py
import zarr
from .h5_ref_to_zarr_attr import h5_ref_to_zarr_attr
from .attr_conversion import h5_to_zarr_attr
from ._util import _is_numeric_dtype


def create_zarr_dataset_from_h5_data(
    zarr_parent_group: zarr.Group,
    h5_shape: Tuple,
    h5_dtype: Any,
    h5_data: Union[Any, None],
    h5f: Union[h5py.File, None],
    name: str,
    label: str,
    h5_chunks: Union[Tuple, None],
    zarr_compressor: Union[Codec, Literal['default'], None] = 'default'
):
    """Create a zarr dataset from an h5py dataset.

    Parameters
    ----------
    zarr_parent_group : zarr.Group
        The parent group in the zarr hierarchy. The new dataset will be created
        in this group.
    h5_shape : tuple
        The shape of the h5py dataset.
    h5_dtype : numpy.dtype
        The dtype of the h5py dataset.
    h5_data : any
        The data of the h5py dataset. If None, the dataset will be created
        without data.
    h5f : h5py.File
        The file that the h5py dataset is in.
    name : str
        The name of the new dataset in the zarr hierarchy.
    label : str
        The name of the h5py dataset for error messages.
    h5_chunks : tuple
        The chunk shape of the h5py dataset.
    zarr_compressor : numcodecs.abc.Codec, 'default', or None
        The codec compressor to use when writing the dataset. If default, the
        default compressor will be used. When None, no compressor will be used.
    """
    if h5_dtype is None:
        raise Exception(f'No dtype in h5_to_zarr_dataset_prep for dataset {label}')
    if len(h5_shape) == 0:
        # scalar dataset
        # zarr doesn't support scalar datasets, so we make an array of shape (1,)
        # and the _SCALAR attribute will be set to True elsewhere to indicate that
        # it is a scalar dataset

        if h5_data is None:
            raise Exception(f'Data must be provided for scalar dataset {label}')

        if zarr_compressor != 'default' and zarr_compressor is not None:
            raise Exception('zarr_compressor is not supported for scalar datasets')

        if np.issubdtype(h5_dtype, np.complexfloating):
            raise Exception(f'Complex scalar datasets are not supported: dataset {label} with dtype {h5_dtype}')

        if _is_numeric_dtype(h5_dtype) or h5_dtype in [bool, np.bool_]:
            # Handle the simple numeric types
            ds = zarr_parent_group.create_dataset(
                name,
                shape=(1,),
                chunks=(1,),
                data=[h5_data[()]] if isinstance(h5_data, h5py.Dataset) or isinstance(h5_data, np.ndarray) else [h5_data],
            )
            ds.attrs['_SCALAR'] = True
            return ds
        elif h5_dtype.kind == 'O':
            # For type object, we are going to use the JSON codec
            # for encoding [scalar_value]
            scalar_value = h5_data[()] if isinstance(h5_data, h5py.Dataset) or isinstance(h5_data, np.ndarray) else h5_data
            if isinstance(scalar_value, (bytes, str)):
                if isinstance(scalar_value, bytes):
                    scalar_value = scalar_value.decode()
                ds = zarr_parent_group.create_dataset(
                    name,
                    shape=(1,),
                    chunks=(1,),
                    data=[scalar_value]
                )
                ds.attrs['_SCALAR'] = True
                return ds
            else:
                raise Exception(f'Unsupported scalar value type: {type(scalar_value)}')
        elif h5_dtype.kind == 'S' or h5_dtype.kind == 'U':
            # byte string
            if h5_data is None:
                raise Exception(f'Data must be provided for scalar dataset {label}')
            scalar_value = h5_data[()] if isinstance(h5_data, h5py.Dataset) or isinstance(h5_data, np.ndarray) else h5_data
            ds = zarr_parent_group.create_dataset(
                name,
                shape=(1,),
                chunks=(1,),
                data=[scalar_value]
            )
            ds.attrs['_SCALAR'] = True
            return ds
        else:
            raise Exception(f'Cannot handle scalar dataset {label} with dtype {h5_dtype}')
    else:
        # not a scalar dataset

        if isinstance(h5_data, list):
            # If we have a list, then we need to convert it to an array
            h5_data = np.array(h5_data)

        if np.issubdtype(h5_dtype, np.complexfloating):
            raise Exception(f'Complex datasets are not supported: dataset {label} with dtype {h5_dtype}')

        if _is_numeric_dtype(h5_dtype) or h5_dtype in [bool, np.bool_]:  # integer, unsigned integer, float, bool
            # This is the normal case of a chunked dataset with a numeric (or boolean) dtype
            if h5_chunks is None:
                # # We require that chunks be specified when writing a dataset with more
                # # than 1 million elements. This is because zarr may default to
                # # suboptimal chunking. Note that the default for h5py is to use the
                # # entire dataset as a single chunk.
                # total_size = int(np.prod(h5_shape)) if len(h5_shape) > 0 else 1
                # if total_size > 1000 * 1000:
                #     raise Exception(f'Chunks must be specified explicitly when writing dataset of shape {h5_shape}')
                h5_chunks = _get_default_chunks(h5_shape, h5_dtype)
            # Note that we are not using the same filters as in the h5py dataset
            return zarr_parent_group.create_dataset(
                name,
                shape=h5_shape,
                chunks=h5_chunks,
                dtype=h5_dtype,
                data=h5_data,
                compressor=zarr_compressor
            )
        elif h5_dtype.kind == 'O':
            # For type object, we are going to use the JSON codec
            if zarr_compressor != 'default' and zarr_compressor is not None:
                raise Exception('zarr_compressor is not supported for object datasets')
            if h5_data is not None:
                if isinstance(h5_data, h5py.Dataset):
                    h5_data = h5_data[:]
                zarr_data = h5_object_data_to_zarr_data(h5_data, h5f=h5f, label=label)
            else:
                zarr_data = None
            object_codec = numcodecs.JSON()
            return zarr_parent_group.create_dataset(
                name,
                shape=h5_shape,
                chunks=h5_chunks,
                dtype=h5_dtype,
                data=zarr_data,
                object_codec=object_codec
            )
        elif h5_dtype.kind == 'S':  # byte string
            if zarr_compressor != 'default' and zarr_compressor is not None:
                raise Exception('zarr_compressor is not supported for byte string datasets')
            if h5_data is None:
                raise Exception(f'Data must be provided when converting dataset {label} with dtype {h5_dtype}')
            return zarr_parent_group.create_dataset(
                name,
                shape=h5_shape,
                chunks=h5_chunks,
                dtype=h5_dtype,
                data=h5_data
            )
        elif h5_dtype.kind == 'U':  # unicode string
            if zarr_compressor != 'default' and zarr_compressor is not None:
                raise Exception('zarr_compressor is not supported for unicode string datasets')
            raise Exception(f'Array of unicode strings not supported: dataset {label} with dtype {h5_dtype} and shape {h5_shape}')
        elif h5_dtype.kind == 'V' and h5_dtype.fields is not None:  # compound dtype
            if zarr_compressor != 'default' and zarr_compressor is not None:
                raise Exception('zarr_compressor is not supported for compound datasets')
            if h5_data is None:
                raise Exception(f'Data must be provided when converting compound dataset {label}')
            h5_data_1d_view = h5_data.ravel()
            zarr_data = np.empty(h5_shape, dtype='object')
            zarr_data_1d_view = zarr_data.ravel()
            for i in range(len(h5_data_1d_view)):
                elmt = tuple([
                    _make_json_serializable(
                        h5_data_1d_view[i][field_name],
                        h5_dtype[field_name],
                        label=f'{label}[{i}].{field_name}',
                        h5f=h5f
                    )
                    for field_name in h5_dtype.names
                ])
                zarr_data_1d_view[i] = elmt
            ds = zarr_parent_group.create_dataset(
                name,
                shape=h5_shape,
                chunks=h5_chunks,
                dtype='object',
                data=zarr_data,
                object_codec=numcodecs.JSON()
            )
            compound_dtype = []
            for name in h5_dtype.names:
                tt = h5_dtype[name]
                if tt == h5py.special_dtype(ref=h5py.Reference):
                    tt = "<REFERENCE>"
                compound_dtype.append((name, str(tt)))
            ds.attrs['_COMPOUND_DTYPE'] = compound_dtype
            return ds
        else:
            raise Exception(f'Not yet implemented (3): dataset {label} with dtype {h5_dtype} and shape {h5_shape}')


@dataclass
class CreateZarrDatasetInfo:
    shape: Tuple
    dtype: Any
    fill_value: Any
    scalar: bool
    compound_dtype: Union[List[Tuple[str, str]], None]


def _make_json_serializable(val: Any, dtype: np.dtype, label: str, h5f: Union[h5py.File, None]) -> Any:
    if dtype.kind in ['i', 'u']:  # integer, unsigned integer
        return int(val)
    elif dtype.kind == 'f':  # float
        return float(val)
    elif dtype.kind == 'b':  # boolean
        return bool(val)
    elif dtype.kind == 'S':  # byte string
        return val.decode()
    elif dtype.kind == 'U':  # unicode string
        return val
    elif dtype == h5py.Reference:
        return h5_to_zarr_attr(val, label=label, h5f=h5f)
    else:
        raise Exception(f'Cannot serialize item {val} with dtype {dtype} when serializing dataset {label} with compound dtype.')


def h5_object_data_to_zarr_data(h5_data: Union[np.ndarray, list], *, h5f: Union[h5py.File, None], label: str) -> np.ndarray:
    from ..LindiH5pyFile.LindiH5pyReference import LindiH5pyReference  # Avoid circular import
    if isinstance(h5_data, list):
        h5_data = np.array(h5_data)
    zarr_data = np.empty(h5_data.shape, dtype='object')
    h5_data_1d_view = h5_data.ravel()
    zarr_data_1d_view = zarr_data.ravel()
    for i, val in enumerate(h5_data_1d_view):
        if isinstance(val, bytes):
            zarr_data_1d_view[i] = val.decode()
        elif isinstance(val, str):
            zarr_data_1d_view[i] = val
        elif isinstance(val, LindiH5pyReference):
            zarr_data_1d_view[i] = {
                '_REFERENCE': val._obj
            }
        elif isinstance(val, h5py.Reference):
            if h5f is None:
                raise Exception(f'h5f cannot be None when converting h5py.Reference to zarr attribute at {label}')
            zarr_data_1d_view[i] = h5_ref_to_zarr_attr(val, h5f=h5f)
        else:
            raise Exception(f'Cannot handle value of type {type(val)} in dataset {label} with dtype {h5_data.dtype} and shape {h5_data.shape}')
    return zarr_data


def _get_default_chunks(shape: Tuple, dtype: Any) -> Tuple:
    dtype_size = np.dtype(dtype).itemsize
    shape_prod_0 = np.prod(shape[1:])
    optimal_chunk_size_bytes = 1024 * 1024 * 20  # 20 MB
    optimal_chunk_size = optimal_chunk_size_bytes // (dtype_size * shape_prod_0)
    if optimal_chunk_size <= shape[0]:
        return shape
    if optimal_chunk_size < 1:
        return (1,) + shape[1:]
    return (optimal_chunk_size,) + shape[1:]



import numpy as np
from zarr.storage import Store as ZarrStore
from ..LindiH5pyFile.LindiReferenceFileSystemStore import LindiReferenceFileSystemStore
from .lindi_tar import LindiTarFile


class LindiTarStore(ZarrStore):
    def __init__(self, *, base_store: LindiReferenceFileSystemStore, tar_file: LindiTarFile):
        self._base_store = base_store
        self._tar_file = tar_file

    def __getitem__(self, key: str):
        return self._base_store.__getitem__(key)

    def __setitem__(self, key: str, value: bytes):
        self.setitems({key: value})

    def setitems(self, items_dict: dict):
        for key, value in items_dict.items():
            key_parts = key.split("/")
            key_base_name = key_parts[-1]

            files_to_write_to_tar = {}

            if key_base_name.startswith('.') or key_base_name.endswith('.json'):  # always inline .zattrs, .zgroup, .zarray, zarr.json
                inline = True
            else:
                # presumably it is a chunk of an array
                if isinstance(value, np.ndarray):
                    value = value.tobytes()
                if not isinstance(value, bytes):
                    print(f"key: {key}, value type: {type(value)}")
                    raise ValueError("Value must be bytes")
                size = len(value)
                inline = size < 1000  # this should be a configurable threshold
            if inline:
                # If inline, save in memory
                return self._base_store.__setitem__(key, value)
            else:
                # If not inline, save it as a new file in the tar file
                key_without_initial_slash = key if not key.startswith("/") else key[1:]
                fname_in_tar = f'blobs/{key_without_initial_slash}'
                if self._tar_file.has_file_with_name(fname_in_tar):
                    v = 2
                    while self._tar_file.has_file_with_name(f'{fname_in_tar}.v{v}'):
                        v += 1
                    fname_in_tar = f'{fname_in_tar}.v{v}'
                files_to_write_to_tar[fname_in_tar] = value

                self._set_ref_reference(key_without_initial_slash, f'./{fname_in_tar}', 0, len(value))

            self._tar_file.write_files(files_to_write_to_tar)

    def __delitem__(self, key: str):
        # We don't actually delete the file from the tar, but maybe it would be
        # smart to put it in .trash in the future
        return self._base_store.__delitem__(key)

    def __iter__(self):
        return self._base_store.__iter__()

    def __len__(self):
        return self._base_store.__len__()

    # These methods are overridden from BaseStore
    def is_readable(self):
        return True

    def is_writeable(self):
        return True

    def is_listable(self):
        return True

    def is_erasable(self):
        return False

    def _set_ref_reference(self, key: str, filename: str, offset: int, size: int):
        rfs = self._base_store.rfs
        if 'refs' not in rfs:
            # this shouldn't happen, but we'll be defensive
            rfs['refs'] = {}
        rfs['refs'][key] = [
            filename,
            offset,
            size
        ]



import os
import json
import random
import urllib.request
from .create_tar_header import create_tar_header


TAR_ENTRY_JSON_SIZE = 1024
INITIAL_TAR_INDEX_JSON_SIZE = 1024 * 8
INITIAL_LINDI_JSON_SIZE = 1024 * 8


class LindiTarFile:
    def __init__(self, tar_path_or_url: str, dir_representation=False):
        self._tar_path_or_url = tar_path_or_url
        self._dir_representation = dir_representation
        self._is_remote = tar_path_or_url.startswith("http://") or tar_path_or_url.startswith("https://")

        if not dir_representation:
            # Load the entry json
            entry_json = _load_bytes_from_local_or_remote_file(self._tar_path_or_url, 512, 512 + TAR_ENTRY_JSON_SIZE)
            entry = json.loads(entry_json)
            index_info = entry['index']

            # Load the index json
            index_json = _load_bytes_from_local_or_remote_file(self._tar_path_or_url, index_info['d'], index_info['d'] + index_info['s'])
            self._index = json.loads(index_json)
            self._index_has_changed = False

            self._index_lookup = {}
            for file in self._index['files']:
                self._index_lookup[file['n']] = file

            # Verify that the index file correctly has a reference to itself
            index_info_2 = self._index_lookup.get(".tar_index.json", None)
            if index_info_2 is None:
                raise ValueError("File .tar_index.json not found in index")
            for k in ['n', 'o', 'd', 's']:
                if k not in index_info_2:
                    raise ValueError(f"File .tar_index.json does not have key {k}")
                if k not in index_info:
                    raise ValueError(f"File .tar_index.json does not have key {k}")
                if index_info_2[k] != index_info[k]:
                    raise ValueError(f"File .tar_index.json has unexpected value for key {k}")
                # Verify that the index file correctly as a reference to the entry file
                entry_info = self._index_lookup.get(".tar_entry.json", None)
                if entry_info is None:
                    raise ValueError("File .tar_entry.json not found in index")
                if entry_info['n'] != ".tar_entry.json":
                    raise ValueError("File .tar_entry.json has unexpected name")
                if entry_info['o'] != 0:
                    raise ValueError("File .tar_entry.json has unexpected offset")
                if entry_info['d'] != 512:
                    raise ValueError("File .tar_entry.json has unexpected data offset")
                if entry_info['s'] != TAR_ENTRY_JSON_SIZE:
                    raise ValueError("File .tar_entry.json has unexpected size")
            self._file = open(self._tar_path_or_url, "r+b") if not self._is_remote else None
        else:
            self._index = None
            self._index_has_changed = False
            self._index_lookup = None
            self._file = None

    def close(self):
        if not self._dir_representation:
            self._update_index_in_file()
            if self._file is not None:
                self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def get_file_info(self, file_name: str):
        if self._dir_representation:
            raise ValueError("Cannot get file info in a directory representation")
        assert self._index_lookup
        return self._index_lookup.get(file_name, None)

    def overwrite_file_content(self, file_name: str, data: bytes):
        if self._is_remote:
            raise ValueError("Cannot overwrite file content in a remote tar file")
        if not self._dir_representation:
            if self._file is None:
                raise ValueError("File is not open")
            info = self.get_file_info(file_name)
            if info is None:
                raise FileNotFoundError(f"File {file_name} not found")
            if info['s'] != len(data):
                raise ValueError("Unexpected problem in overwrite_file_content(): data size must match the size of the existing file")
            self._file.seek(info['d'])
            self._file.write(data)
        else:
            # for safety:
            file_parts = file_name.split("/")
            for part in file_parts[:-1]:
                if part.startswith('..'):
                    raise ValueError(f"Invalid path: {file_name}")
            fname = self._tar_path_or_url + "/" + file_name
            with open(fname, "wb") as f:
                f.write(data)

    def trash_file(self, file_name: str):
        if self._is_remote:
            raise ValueError("Cannot trash a file in a remote tar file")
        if not self._dir_representation:
            if self._file is None:
                raise ValueError("File is not open")
            assert self._index
            assert self._index_lookup
            info = self.get_file_info(file_name)
            if info is None:
                raise FileNotFoundError(f"File {file_name} not found")
            zeros = b"-" * info['s']
            self._file.seek(info['d'])
            self._file.write(zeros)
            self._change_name_of_file(
                file_name,
                f'.trash/{file_name}.{_create_random_string()}'
            )
            self._index['files'] = [file for file in self._index['files'] if file['n'] != file_name]
            del self._index_lookup[file_name]
            self._index_has_changed = True
        else:
            # for safety:
            file_parts = file_name.split("/")
            for part in file_parts[:-1]:
                if part.startswith('..'):
                    raise ValueError(f"Invalid path: {file_name}")
            fname = self._tar_path_or_url + "/" + file_name
            os.remove(fname)

    def write_rfs(self, rfs: dict):
        if self._is_remote:
            raise ValueError("Cannot write a file in a remote tar file")

        rfs_json = json.dumps(rfs, indent=2, sort_keys=True)

        if not self._dir_representation:
            existing_lindi_json_info = self.get_file_info("lindi.json")
            if existing_lindi_json_info is not None:
                file_size = existing_lindi_json_info['s']
                if file_size >= len(rfs_json):
                    # We are going to overwrite the existing lindi.json with the new
                    # one. But first we pad it with spaces to the same size as the
                    # existing one.
                    padding = b" " * (file_size - len(rfs_json))
                    rfs_json = rfs_json.encode() + padding
                    self.overwrite_file_content("lindi.json", rfs_json)
                else:
                    # In this case we need to trash the existing file and write a new one
                    # at the end of the tar file.
                    self.trash_file("lindi.json")
                    rfs_json = _pad_bytes_to_leave_room_for_growth(rfs_json, INITIAL_LINDI_JSON_SIZE)
                    self.write_file("lindi.json", rfs_json)
            else:
                # We are writing a new lindi.json.
                rfs_json = _pad_bytes_to_leave_room_for_growth(rfs_json, INITIAL_LINDI_JSON_SIZE)
                self.write_file("lindi.json", rfs_json)
        else:
            with open(self._tar_path_or_url + "/lindi.json", "wb") as f:
                f.write(rfs_json.encode())

    def get_file_byte_range(self, file_name: str) -> tuple:
        if self._dir_representation:
            raise ValueError("Cannot get file byte range in a directory representation")
        info = self.get_file_info(file_name)
        if info is None:
            raise FileNotFoundError(f"File {file_name} not found in tar file")
        return info['d'], info['d'] + info['s']

    def has_file_with_name(self, file_name: str) -> bool:
        if not self._dir_representation:
            return self.get_file_info(file_name) is not None
        else:
            return os.path.exists(self._tar_path_or_url + "/" + file_name)

    def _change_name_of_file(self, file_name: str, new_file_name: str):
        if self._dir_representation:
            raise ValueError("Cannot change the name of a file in a directory representation")
        if self._is_remote:
            raise ValueError("Cannot change the name of a file in a remote tar file")
        if self._file is None:
            raise ValueError("File is not open")
        info = self.get_file_info(file_name)
        if info is None:
            raise FileNotFoundError(f"File {file_name} not found")
        header_start_byte = info['o']
        file_name_byte_range = (header_start_byte + 0, header_start_byte + 100)
        file_name_prefix_byte_range = (header_start_byte + 345, header_start_byte + 345 + 155)
        self._file.seek(file_name_byte_range[0])
        self._file.write(new_file_name.encode())
        # set the rest of the field to zeros
        self._file.write(b"\x00" * (file_name_byte_range[1] - file_name_byte_range[0] - len(new_file_name)))

        self._file.seek(file_name_prefix_byte_range[0])
        self._file.write(b"\x00" * (file_name_prefix_byte_range[1] - file_name_prefix_byte_range[0]))

        _fix_checksum_in_header(self._file, header_start_byte)

    def write_file(self, file_name: str, data: bytes):
        self.write_files({file_name: data})

    def write_files(self, files: dict):
        if self._is_remote:
            raise ValueError("Cannot write a file in a remote tar file")
        if not self._dir_representation:
            assert self._index
            assert self._index_lookup
            if self._file is None:
                raise ValueError("File is not open")
            self._file.seek(-1024, 2)
            hh = self._file.read(1024)
            if hh != b"\x00" * 1024:
                raise ValueError("The tar file does not end with 1024 bytes of zeros")
            self._file.seek(-1024, 2)

            file_pos = self._file.tell()

            for file_name, data in files.items():
                x = {
                    'n': file_name,
                    'o': file_pos,
                    'd': file_pos + 512,  # we assume the header is 512 bytes
                    's': len(data)
                }

                # write the tar header
                tar_header = create_tar_header(file_name, len(data))

                # pad up to blocks of 512
                if len(data) % 512 != 0:
                    padding_len = 512 - len(data) % 512
                else:
                    padding_len = 0

                self._file.write(tar_header)
                self._file.write(data)
                self._file.write(b"\x00" * padding_len)
                file_pos += 512 + len(data) + padding_len

                self._index['files'].append(x)
                self._index_lookup[file_name] = x
                self._index_has_changed = True

            # write the 1024 bytes marking the end of the file
            self._file.write(b"\x00" * 1024)
        else:
            for file_name, data in files.items():
                # for safety:
                file_parts = file_name.split("/")
                for part in file_parts[:-1]:
                    if part.startswith('..'):
                        raise ValueError(f"Invalid path: {file_name}")
                fname = self._tar_path_or_url + "/" + file_name
                parent_dir = os.path.dirname(fname)
                if not os.path.exists(parent_dir):
                    os.makedirs(parent_dir)
                with open(fname, "wb") as f:
                    f.write(data)

    def read_file(self, file_name: str) -> bytes:
        if not self._dir_representation:
            info = self.get_file_info(file_name)
            if info is None:
                raise FileNotFoundError(f"File {file_name} not found for {self._tar_path_or_url}")
            start_byte = info['d']
            size = info['s']
            return _load_bytes_from_local_or_remote_file(self._tar_path_or_url, start_byte, start_byte + size)
        else:
            return _load_all_bytes_from_local_or_remote_file(self._tar_path_or_url + "/" + file_name)

    @staticmethod
    def create(fname: str, *, rfs: dict, dir_representation=False):
        if not dir_representation:
            with open(fname, "wb") as f:
                # Define the sizes and names of the entry and index files
                tar_entry_json_name = ".tar_entry.json"
                tar_entry_json_size = TAR_ENTRY_JSON_SIZE
                tar_index_json_size = INITIAL_TAR_INDEX_JSON_SIZE
                tar_index_json_name = ".tar_index.json"
                tar_index_json_offset = 512 + TAR_ENTRY_JSON_SIZE
                tar_index_json_offset_data = tar_index_json_offset + 512

                # Define the content of .tar_entry.json
                initial_entry_json = json.dumps({
                    'index': {
                        'n': tar_index_json_name,
                        'o': tar_index_json_offset,
                        'd': tar_index_json_offset_data,
                        's': tar_index_json_size
                    }
                }, indent=2, sort_keys=True)
                initial_entry_json = initial_entry_json.encode() + b" " * (tar_entry_json_size - len(initial_entry_json))

                # Define the content of .tar_index.json
                initial_index_json = json.dumps({
                    'files': [
                        {
                            'n': tar_entry_json_name,
                            'o': 0,
                            'd': 512,
                            's': tar_entry_json_size
                        },
                        {
                            'n': tar_index_json_name,
                            'o': tar_index_json_offset,
                            'd': tar_index_json_offset_data,
                            's': tar_index_json_size
                        }
                    ]
                }, indent=2, sort_keys=True)
                initial_index_json = initial_index_json.encode() + b" " * (tar_index_json_size - len(initial_index_json))

                # Write the initial entry file (.tar_entry.json). This will always
                # be the first file in the tar file, and has a fixed size.
                header = create_tar_header(tar_entry_json_name, tar_entry_json_size)
                f.write(header)
                f.write(initial_entry_json)

                # Write the initial index file (.tar_index.json) this will start as
                # the second file in the tar file but as it grows outside the
                # initial bounds, a new index file will be appended to the end of
                # the tar, and then entry file will be updated accordingly to point
                # to the new index file.
                header = create_tar_header(tar_index_json_name, tar_index_json_size)
                f.write(header)
                f.write(initial_index_json)

                f.write(b"\x00" * 1024)
        else:
            if os.path.exists(fname):
                raise ValueError(f"Directory {fname} already exists")
            os.makedirs(fname)

        # write the rfs file
        tf = LindiTarFile(fname, dir_representation=dir_representation)
        tf.write_rfs(rfs)
        tf.close()

    def _update_index_in_file(self):
        if self._dir_representation:
            return
        assert self._index
        if not self._index_has_changed:
            return
        if self._is_remote:
            raise ValueError("Cannot update the index in a remote tar file")
        if self._file is None:
            raise ValueError("File is not open")
        existing_index_json = self.read_file(".tar_index.json")
        new_index_json = json.dumps(self._index, indent=2, sort_keys=True)
        if len(new_index_json) <= len(existing_index_json):
            # we can overwrite the existing index file
            new_index_json = new_index_json.encode() + b" " * (len(existing_index_json) - len(new_index_json))
            self.overwrite_file_content(".tar_index.json", new_index_json)
        else:
            # we must create a new index file
            self.trash_file(".tar_index.json")

            # after we trash the file, the index has changed once again
            new_index_json = json.dumps(self._index, indent=2, sort_keys=True)
            new_index_json = _pad_bytes_to_leave_room_for_growth(new_index_json, INITIAL_TAR_INDEX_JSON_SIZE)
            new_index_json_size = len(new_index_json)
            self.write_file(".tar_index.json", new_index_json)

            # now the index has changed once again, but we assume it doesn't exceed the size
            new_index_json = json.dumps(self._index, indent=2, sort_keys=True)
            new_index_json = new_index_json.encode() + b" " * (new_index_json_size - len(new_index_json))
            self.overwrite_file_content(".tar_index.json", new_index_json)

            tar_index_info = next(file for file in self._index['files'] if file['n'] == ".tar_index.json")
            new_entry_json = json.dumps({
                'index': {
                    'n': tar_index_info['n'],
                    'o': tar_index_info['o'],
                    'd': tar_index_info['d'],
                    's': tar_index_info['s']
                }
            }, indent=2, sort_keys=True)
            new_entry_json = new_entry_json.encode() + b" " * (TAR_ENTRY_JSON_SIZE - len(new_entry_json))
            self._file.seek(512)
            self._file.write(new_entry_json)
        self._file.flush()
        self._index_has_changed = False


def _download_file_byte_range(url: str, start: int, end: int) -> bytes:
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3",
        "Range": f"bytes={start}-{end - 1}"
    }
    req = urllib.request.Request(url, headers=headers)
    with urllib.request.urlopen(req) as response:
        return response.read()


def _load_all_bytes_from_local_or_remote_file(path_or_url: str) -> bytes:
    if path_or_url.startswith("http://") or path_or_url.startswith("https://"):
        with urllib.request.urlopen(path_or_url) as response:
            return response.read()
    else:
        with open(path_or_url, "rb") as f:
            return f.read()


def _load_bytes_from_local_or_remote_file(path_or_url: str, start: int, end: int) -> bytes:
    if path_or_url.startswith("http://") or path_or_url.startswith("https://"):
        return _download_file_byte_range(path_or_url, start, end)
    else:
        with open(path_or_url, "rb") as f:
            f.seek(start)
            return f.read(end - start)


def _pad_bytes_to_leave_room_for_growth(x: str, initial_size: int) -> bytes:
    total_size = initial_size
    while total_size < len(x) * 4:
        total_size *= 2
    padding = b" " * (total_size - len(x))
    return x.encode() + padding


def _fix_checksum_in_header(f, header_start_byte):
    f.seek(header_start_byte)
    header = f.read(512)

    # From https://en.wikipedia.org/wiki/Tar_(computing)
    # The checksum is calculated by taking the sum of the unsigned byte values
    # of the header record with the eight checksum bytes taken to be ASCII
    # spaces (decimal value 32). It is stored as a six digit octal number with
    # leading zeroes followed by a NUL and then a space. Various implementations
    # do not adhere to this format. In addition, some historic tar
    # implementations treated bytes as signed. Implementations typically
    # calculate the checksum both ways, and treat it as good if either the
    # signed or unsigned sum matches the included checksum.

    header_byte_list = []
    for byte in header:
        header_byte_list.append(byte)
    for i in range(148, 156):
        header_byte_list[i] = 32
    sum = 0
    for byte in header_byte_list:
        sum += byte
    checksum = oct(sum).encode()[2:]
    while len(checksum) < 6:
        checksum = b"0" + checksum
    checksum += b"\x00 "
    f.seek(header_start_byte + 148)
    f.write(checksum)


def _create_random_string():
    return "".join(random.choices("abcdefghijklmnopqrstuvwxyz", k=10))
